/* TTMPConn.c
 *
 * TTMPConnection implementation.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <stdlib.h>
#include "TTMPConn.h"
#include "TTMPMsg.h"
#include "TTMPPriv.h"
#include "TTMPPlat.h"
#include "hexdump.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#define TTMP_PROTO_SIZE (TTMP_FS_Uint32 + TTMP_FS_Int32 + TTMP_FS_Int32)
#define TTMP_HDR_SIZE   (TTMP_FS_Uint32 + TTMP_FS_Uint32)
#define TTMP_MSG_MIN    (TTMP_FS_Uint32 + TTMP_FS_Uint32 + TTMP_FS_Uint32)

/* Define some dirty streaming macros */
#define STREAM(_type, _val)           \
	do {                              \
		*((T##_type *) buf) = (_val); \
		buf += TTMP_FS_##_type;       \
	} while (0)

#define UNSTREAM(_type, _val)         \
	do {                              \
		(_val) = *((T##_type *) buf); \
		buf += TTMP_FS_##_type;       \
	} while (0)

static inline void TTMPBufferInit(TTMPBuffer *buf)
{
	buf->data = NULL;
	buf->size = 0;
	buf->ofs = 0;
}

static inline void TTMPBufferDestroy(TTMPBuffer *buf)
{
	if (buf->data != NULL)
		free(buf->data);
	TTMPBufferInit(buf);
}

int TTMPConnectionCreateInternal(TTMPConnection *conn, TInt32 timeout, TInt32 major, TInt32 minor, TTMPSocket sock)
{
	TTMPConnection newconn;

	if (conn == NULL)
		return TTMP_ERR_CONNULL;
	CHECK_TIMEOUT(timeout);

	newconn = (TTMPConnection) malloc(sizeof(struct sTTMPConnection));
	if (newconn == NULL)
		return TTMP_ERR_NOMEMORY;

	newconn->magic = TTMP_CON_MAGIC;
	newconn->timeout = timeout;
	newconn->major = major;
	newconn->minor = minor;
	newconn->sendstate = SS_SILENT;
	newconn->recvstate = RS_SILENT;
	newconn->socket = sock;
	TTMPBufferInit(&newconn->sendbuf);
	TTMPBufferInit(&newconn->recvbuf);

	/* Return connection object to caller */
	*conn = newconn;

	return TTMP_ERR_NONE;
}

int TTMPConnectionCreate(TTMPConnection *conn, TInt32 timeout, TInt32 major, TInt32 minor)
{
	return TTMPConnectionCreateInternal(conn, timeout, major, minor, INVALID_SOCKET);
}

int TTMPConnectionDestroy(TTMPConnection conn)
{
	int res;

	CHECK_OBJ(CON, conn);

	/* Close connection if needed */
	res = TTMPConnectionClose(conn);

	/* Zap magic to provoke errors */
	conn->magic = 0;

	/* Delete connection struct */
	free(conn);

	return res;
}

int TTMPConnectionValidate(TTMPConnection conn)
{
	CHECK_OBJ(CON, conn);
	return TTMP_ERR_NONE;
}

int TTMPConnectionConnect(TTMPConnection conn, cTTMPAddress addr)
{
	return TTMPConnectionConnectTimeout(conn, addr, TTMP_TO_INFINITE);
}

int TTMPConnectionConnectTimeout(TTMPConnection conn, cTTMPAddress addr, TInt32 timeout)
{
	TTMPSocketAddress sa;
	int res;

	CHECK_OBJ(CON, conn);

	/* We should be disconnected */
	if (conn->sendstate != SS_SILENT || conn->recvstate != RS_SILENT) {
		TTMPDBG("Connecting while already connected\n");
		TTMPConnectionClose(conn);
		return TTMP_ERR_DISCONN;
	}

	/* Parse the address */
	res = TTMPSocketAddressParse(addr, &sa);
	if (res != TTMP_ERR_NONE)
		return res;

	/* Create socket if needed */
	if (SOCKET_INVALID(conn->socket)) {
		res = TTMPSocketCreate(&conn->socket, sa.domain);
		if (res != TTMP_ERR_NONE)
			return res;
	}

	/* Connect socket */
	res = TTMPSocketConnect(conn->socket, &sa, timeout);
	if (res != TTMP_ERR_NONE) {
		TTMPConnectionClose(conn);
		return res;
	}
	conn->sendstate = SS_CONNECTED;
	conn->recvstate = RS_CONNECTED;

	res = TTMPConnectionCheckVersion(conn, timeout);
	if (res != TTMP_ERR_NONE) {
		TTMPConnectionClose(conn);
		return res;
	}

	return TTMP_ERR_NONE;
}

int TTMPConnectionCheckVersion(TTMPConnection conn, TInt32 timeout)
{
	int res;
	TUint8 protobuf[TTMP_PROTO_SIZE];
	TUint8 *buf;
	TUint32 peerproto;
	TInt32 peermajor;
	TInt32 peerminor;

	CHECK_OBJ(CON, conn);

	/* We should just have connected */
	if (conn->sendstate != SS_CONNECTED || conn->recvstate != RS_CONNECTED) {
		TTMPERR("Checking version while not just connected\n");
		return TTMP_ERR_INTERNAL;
	}

	/* Announce versions to peer */
	conn->sendbuf.data = protobuf;
	conn->sendbuf.size = TTMP_PROTO_SIZE;
	conn->sendbuf.ofs = 0;

	buf = protobuf;
	STREAM(Uint32, TTMP_PROTO_VER);
	STREAM(Int32, conn->major);
	STREAM(Int32, conn->minor);

	res = TTMPSocketSendBuffer(conn->socket, &conn->sendbuf, timeout);
	TTMPBufferInit(&conn->sendbuf);
	if (res != TTMP_ERR_NONE) {
		TTMPMSG("send version failed, %d/%d\n", res, socket_errno);
		return res;
	}

	/* Read versions from peer */
	conn->recvbuf.data = protobuf;
	conn->recvbuf.size = TTMP_PROTO_SIZE;
	conn->recvbuf.ofs = 0;

	res = TTMPSocketReceiveBuffer(conn->socket, &conn->recvbuf, timeout);
	TTMPBufferInit(&conn->recvbuf);
	if (res != TTMP_ERR_NONE) {
		TTMPMSG("recv version failed, %d/%d\n", res, socket_errno);
		return res;
	}


	buf = protobuf;
	UNSTREAM(Uint32, peerproto);
	UNSTREAM(Int32, peermajor);
	UNSTREAM(Int32, peerminor);

	if (peerproto > TTMP_PROTO_VER) {
		TTMPMSG("peerproto %u > %u\n", peerproto, TTMP_PROTO_VER);
		return TTMP_ERR_PROTOVER;
	}

	if (peermajor != conn->major || peerminor != conn->minor) {
		TTMPMSG("peerver %d.%d != %d.%d\n", peermajor, peerminor, conn->major, conn->minor);
		return TTMP_ERR_INCOMPAT;
	}

	conn->sendstate = SS_SLEEPING;
	conn->recvstate = RS_SLEEPING;
	return TTMP_ERR_NONE;
}

int TTMPConnectionSendTimeout(TTMPConnection conn, cTTMPMessage msg, TInt32 timeout)
{
	int res;

	CHECK_OBJ(CON, conn);
	CHECK_OBJ(MSG, msg);
	CHECK_TIMEOUT(timeout);

	switch (conn->sendstate) {
	case SS_SILENT:
		/* We're not connected yet */
		TTMPDBG("Sending while not connected\n");
		return TTMP_ERR_NOTCONN;
	case SS_SLEEPING:
		/* Start sending the message */
		TTMPDBG("Start sending message\n");
		res = TTMPMessageMarshal(msg, &conn->sendbuf.data, &conn->sendbuf.size);
		if (res != TTMP_ERR_NONE)
			return res;
#ifdef DEBUG
		TTMPDBG("Marshalled message:\n");
		hexdump(stderr, conn->sendbuf.data, conn->sendbuf.size);
#endif
		conn->sendbuf.ofs = 0;
		conn->sendstate = SS_MESSAGE;
		/* FALLTHROUGH */
	case SS_MESSAGE:
		/* Continue sending message */
		TTMPDBG("Continue sending message\n");
		res = TTMPSocketSendBuffer(conn->socket, &conn->sendbuf, timeout);
		if (res == TTMP_ERR_TIMEOUT)
			return res;
		if (res != TTMP_ERR_NONE)
			break;
		/* Message completely sent */
		conn->sendstate = SS_SLEEPING;
		res = TTMP_ERR_NONE;
		break;
	default:
		/* Shouldn't happen */
		TTMPERR("Unexpected send state %d\n", conn->sendstate);
		return TTMP_ERR_INTERNAL;
	}

	/* Clean up send buffer and reset state */
	TTMPBufferDestroy(&conn->sendbuf);
	conn->sendstate = SS_SLEEPING;
	return res;
}

int TTMPConnectionSend(TTMPConnection conn, cTTMPMessage msg)
{
	CHECK_OBJ(CON, conn);
	return TTMPConnectionSendTimeout(conn, msg, conn->timeout);
}

int TTMPConnectionSendBlocking(TTMPConnection conn, cTTMPMessage msg)
{
	CHECK_OBJ(CON, conn);
	return TTMPConnectionSendTimeout(conn, msg, TTMP_TO_INFINITE);
}

int TTMPConnectionReceiveTimeout(TTMPConnection conn, TTMPMessage *msg, TInt32 timeout)
{
	int res;
	TUint32 magic;
	TUint32 size;
	TUint8 *buf;
	TTMPMessage newmsg;

	CHECK_OBJ(CON, conn);
	if (msg == NULL)
		return TTMP_ERR_ARGUMENT;
	CHECK_TIMEOUT(timeout);

	switch (conn->recvstate) {
	case RS_SILENT:
		/* We're not connected yet */
		TTMPDBG("Receiving while not connected\n");
		return TTMP_ERR_NOTCONN;
	case RS_SLEEPING:
		/* Start receiving the message header */
		TTMPDBG("Start receiving message header\n");
		conn->recvbuf.data = malloc(TTMP_HDR_SIZE);
		if (conn->recvbuf.data == NULL)
			return TTMP_ERR_NOMEMORY;
		conn->recvbuf.size = TTMP_HDR_SIZE;
		conn->recvbuf.ofs = 0;
		conn->recvstate = RS_HEADER;
		/* FALLTHROUGH */
	case RS_HEADER:
		/* Continue receiving the message header */
		TTMPDBG("Continue receiving message header\n");
		res = TTMPSocketReceiveBuffer(conn->socket, &conn->recvbuf, timeout);
		if (res == TTMP_ERR_TIMEOUT)
			return res;
		if (res != TTMP_ERR_NONE)
			break;
		/* Header completely received */
		buf = conn->recvbuf.data;
		UNSTREAM(Uint32, magic);
		UNSTREAM(Uint32, size);
		if (magic != TTMP_MSG_MAGIC || size < TTMP_MSG_MIN) {
			TTMPDBG("Incorrect magic %u or size %u\n", magic, size);
			res = TTMP_ERR_MSGCORR;
			break;
		}
		buf = realloc(conn->recvbuf.data, size);
		if (buf == NULL) {
			TTMPDBG("Unable to realloc to %u\n", size);
			res = TTMP_ERR_NOMEMORY;
			break;
		}
		conn->recvbuf.data = buf;
		conn->recvbuf.size = size;
		/* conn->recvbuf.ofs can stay the same */
		/* Start receiving message contents */
		conn->recvstate = RS_CONTENTS;
		/* FALLTHROUGH */
	case RS_CONTENTS:
		/* Receiving the message contents */
		TTMPDBG("Receiving message contents\n");
		res = TTMPSocketReceiveBuffer(conn->socket, &conn->recvbuf, conn->timeout);
		if (res == TTMP_ERR_TIMEOUT)
			return res;
		if (res != TTMP_ERR_NONE)
			break;
		/* Message completely received */
		conn->recvstate = RS_SLEEPING;
		/* Attempt to unmarshal it */
#ifdef DEBUG
		TTMPDBG("Message before unmarshalling:\n");
		hexdump(stderr, conn->recvbuf.data, conn->recvbuf.size);
#endif
		res = TTMPMessageUnmarshal(conn->recvbuf.data, conn->recvbuf.size, &newmsg);
		if (res != TTMP_ERR_NONE)
			break;
		/* Return message object to caller */
		*msg = newmsg;
		res = TTMP_ERR_NONE;
		break;
	default:
		/* Shouldn't happen */
		TTMPERR("Unexpected recv state %d\n", conn->recvstate);
		return TTMP_ERR_INTERNAL;
	}

	/* Clean up receive buffer and reset state */
	TTMPBufferDestroy(&conn->recvbuf);
	conn->recvstate = RS_SLEEPING;
	return res;
}

int TTMPConnectionReceive(TTMPConnection conn, TTMPMessage *msg)
{
	CHECK_OBJ(CON, conn);
	return TTMPConnectionReceiveTimeout(conn, msg, conn->timeout);
}

int TTMPConnectionReceiveBlocking(TTMPConnection conn, TTMPMessage *msg)
{
	CHECK_OBJ(CON, conn);
	return TTMPConnectionReceiveTimeout(conn, msg, TTMP_TO_INFINITE);
}

int TTMPConnectionReceiveMultiTimeout(TTMPConnection conn[], TUint32 num, TInt32 timeout, TTMPMessage *msg, TUint32 *index)
{
	TUint32 i;

	if (conn == NULL || num <= 0 || msg == NULL || index == NULL)
		return TTMP_ERR_ARGUMENT;
	CHECK_TIMEOUT(timeout);
	for (i = 0; i < num; ++i)
		CHECK_OBJ(CON, conn[i]);

	/* Determine state for each connection */

	return TTMP_ERR_NOTIMPL;
}

int TTMPConnectionReceiveMultiBlocking(TTMPConnection conn[], TUint32 num, TTMPMessage *msg, TUint32 *index)
{
	return TTMPConnectionReceiveMultiTimeout(conn, num, TTMP_TO_INFINITE, msg, index);
}

int TTMPConnectionClose(TTMPConnection conn)
{
	int res;

	CHECK_OBJ(CON, conn);

	if (SOCKET_VALID(conn->socket)) {
		res = TTMPSocketDestroy(conn->socket);
		conn->socket = INVALID_SOCKET;
	} else {
		res = TTMP_ERR_NONE;
	}
	conn->sendstate = SS_SILENT;
	conn->recvstate = RS_SILENT;

	/* Clean out buffers */
	TTMPBufferDestroy(&conn->sendbuf);
	TTMPBufferDestroy(&conn->recvbuf);

	return res;
}

/* EOF */
