/* TTMPSock.c
 *
 * Common TTMPSocket implementation.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include "TTMPSock.h"
#include "TTMPPriv.h"
#include "TTMPPlat.h"

int TTMPSocketCreate(TTMPSocket *outsock, int domain)
{
	TTMPSocket sock;
	int res;

	if (outsock == NULL)
		return TTMP_ERR_ARGUMENT;

	sock = socket(domain, SOCK_STREAM, 0);
	if (SOCKET_INVALID(sock))
		return TTMP_ERR_SYSTEM;

	res = TTMPSocketNonBlocking(sock);
	if (res != TTMP_ERR_NONE) {
		/* Don't leak sockets in case of errors */
		closesocket(sock);
		return res;
	}

	*outsock = sock;
	return TTMP_ERR_NONE;
}

int TTMPSocketDestroy(TTMPSocket sock)
{
	if (SOCKET_INVALID(sock))
		return TTMP_ERR_ARGUMENT;

	if (closesocket(sock) == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	return TTMP_ERR_NONE;
}

int TTMPSocketSelect(TTMPSocket sock, fd_set *rset, fd_set *wset, fd_set *eset, TInt32 timeout)
{
	struct timeval *ptv;
	struct timeval tv;
	int res;

	if (timeout < 0) {
		/* Blocking */
		ptv = NULL;
	} else {
		if (timeout < 1000000) {
			tv.tv_sec = 0;
			tv.tv_usec = timeout;
		} else {
			tv.tv_sec = timeout / 1000000;
			tv.tv_usec = timeout % 1000000;
		}
		ptv = &tv;
	}
	res = select((int) sock + 1, rset, wset, eset, ptv);
	if (res == SOCKET_ERROR) {
		TTMPDBG("select failed, errno %d\n", socket_errno);
		return TTMP_ERR_SYSTEM;
	} else if (res == 0) {
		TTMPDBG("select timed out\n");
		return TTMP_ERR_TIMEOUT;
	} else if (res > 1) {
		/* Shouldn't happen */
		TTMPERR("select returns %d\n", res);
		return TTMP_ERR_INTERNAL;
	}

	return TTMP_ERR_NONE;
}

int TTMPSocketConnectSync(TTMPSocket sock, const TTMPSocketAddress *sa)
{
	return TTMPSocketConnect(sock, sa, TTMP_TO_INFINITE);
}

int TTMPSocketBind(TTMPSocket sock, const TTMPSocketAddress *sa)
{
	if (SOCKET_INVALID(sock) || sa == NULL || sa->domain == INVALID_DOMAIN)
		return TTMP_ERR_ARGUMENT;

	TTMPSocketUnlink(sa);
	if (bind(sock, (struct sockaddr *) &sa->addr, sa->size) == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	return TTMP_ERR_NONE;
}

int TTMPSocketUnbind(TTMPSocket sock, const TTMPSocketAddress *sa)
{
	if (SOCKET_INVALID(sock) || sa == NULL || sa->domain == INVALID_DOMAIN)
		return TTMP_ERR_ARGUMENT;

	TTMPSocketUnlink(sa);
	return TTMP_ERR_NONE;
}

int TTMPSocketListen(TTMPSocket sock, int maxconn)
{
	if (SOCKET_INVALID(sock))
		return TTMP_ERR_ARGUMENT;

	if (listen(sock, maxconn) == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	return TTMP_ERR_NONE;
}

int TTMPSocketAcceptSync(TTMPSocket sock, TTMPSocket *newsock)
{
	return TTMPSocketAccept(sock, newsock, TTMP_TO_INFINITE);
}

int TTMPSocketAccept(TTMPSocket sock, TTMPSocket *newsock, TInt32 timeout)
{
	fd_set rset;
	int res;
	TTMPSocket s;

	if (SOCKET_INVALID(sock) || newsock == NULL)
		return TTMP_ERR_ARGUMENT;

	FD_ZERO(&rset);
	FD_SET(sock, &rset);
	res = TTMPSocketSelect(sock, &rset, NULL, NULL, timeout);
	if (res != TTMP_ERR_NONE)
		return res;

	if (!FD_ISSET(sock, &rset)) {
		/* Shouldn't happen */
		TTMPERR("socket not set in read fdset\n");
		return TTMP_ERR_INTERNAL;
	}

	/* Socket is now accept'able */
	s = accept(sock, NULL, NULL);
	if (s == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	*newsock = s;
	return TTMP_ERR_NONE;
}

int TTMPSocketSendSync(TTMPSocket sock, const TUint8 *buf, TUint32 len, TUint32 *sent)
{
	return TTMPSocketSend(sock, buf, len, sent, TTMP_TO_INFINITE);
}

int TTMPSocketSend(TTMPSocket sock, const TUint8 *buf, TUint32 len, TUint32 *sent, TInt32 timeout)
{
	fd_set wset;
	int res;
	ssize_t ns;

	if (SOCKET_INVALID(sock) || buf == NULL || len == 0 || sent == NULL)
		return TTMP_ERR_ARGUMENT;

	FD_ZERO(&wset);
	FD_SET(sock, &wset);
	res = TTMPSocketSelect(sock, NULL, &wset, NULL, timeout);
	if (res != TTMP_ERR_NONE)
		return res;

	if (!FD_ISSET(sock, &wset)) {
		/* Shouldn't happen */
		TTMPERR("socket not set in write fdset\n");
		return TTMP_ERR_INTERNAL;
	}

	/* Socket is now writable */
	ns = send(sock, buf, len, MSG_NOSIGNAL);
	if (ns == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	*sent = (TUint32) ns;
	return TTMP_ERR_NONE;
}

int TTMPSocketSendBuffer(TTMPSocket sock, TTMPBuffer *buf, TInt32 timeout)
{
	TUint32 tosend;
	int res;
	TUint32 sent;

	for (;;) {
		if (buf->ofs > buf->size) {
			/* Shouldn't happen */
			TTMPERR("buffer ofs %u > size %u\n", buf->ofs, buf->size);
			res = TTMP_ERR_INTERNAL;
			break;
		} else if (buf->ofs == buf->size) {
			/* Done */
			res = TTMP_ERR_NONE;
			break;
		}
		tosend = buf->size - buf->ofs;
		res = TTMPSocketSend(sock, buf->data + buf->ofs, tosend, &sent, timeout);
		if (res != TTMP_ERR_NONE) {
			TTMPDBG("send failed\n");
			break;
		} else if (sent > tosend) {
			/* Shouldn't happen */
			TTMPERR("sent %u > tosend %u\n", sent, tosend);
			res = TTMP_ERR_INTERNAL;
			break;
		} else if (sent == 0) {
			TTMPMSG("peer disconnected?\n");
			res = TTMP_ERR_CONNEOF;
			break;
		}
		buf->ofs += sent;
	}

	return res;
}

int TTMPSocketReceiveSync(TTMPSocket sock, TUint8 *buf, TUint32 len, TUint32 *received)
{
	return TTMPSocketReceive(sock, buf, len, received, TTMP_TO_INFINITE);
}

int TTMPSocketReceive(TTMPSocket sock, TUint8 *buf, TUint32 len, TUint32 *received, TInt32 timeout)
{
	fd_set rset;
	int res;
	ssize_t nr;

	if (SOCKET_INVALID(sock) || buf == NULL || len == 0 || received == NULL)
		return TTMP_ERR_ARGUMENT;

	FD_ZERO(&rset);
	FD_SET(sock, &rset);
	res = TTMPSocketSelect(sock, &rset, NULL, NULL, timeout);
	if (res != TTMP_ERR_NONE)
		return res;

	if (!FD_ISSET(sock, &rset)) {
		/* Shouldn't happen */
		TTMPERR("socket not set in read fdset\n");
		return TTMP_ERR_INTERNAL;
	}

	/* Socket is now readable */
	nr = recv(sock, buf, len, MSG_NOSIGNAL);
	if (nr == SOCKET_ERROR)
		return TTMP_ERR_SYSTEM;

	*received = (TUint32) nr;
	return TTMP_ERR_NONE;
}

int TTMPSocketReceiveBuffer(TTMPSocket sock, TTMPBuffer *buf, TInt32 timeout)
{
	TUint32 torecv;
	int res;
	TUint32 received;

	for (;;) {
		if (buf->ofs > buf->size) {
			/* Shouldn't happen */
			TTMPERR("buffer ofs %u > size %u\n", buf->ofs, buf->size);
			res = TTMP_ERR_INTERNAL;
			break;
		} else if (buf->ofs == buf->size) {
			/* Done */
			res = TTMP_ERR_NONE;
			break;
		}
		torecv = buf->size - buf->ofs;
		res = TTMPSocketReceive(sock, buf->data + buf->ofs, torecv, &received, timeout);
		if (res != TTMP_ERR_NONE) {
			TTMPDBG("receive failed\n");
			break;
		} else if (received > torecv) {
			/* Shouldn't happen */
			TTMPERR("received %u > torecv %u\n", received, torecv);
			res = TTMP_ERR_INTERNAL;
			break;
		} else if (received == 0) {
			TTMPDBG("peer disconnected\n");
			res = TTMP_ERR_CONNEOF;
			break;
		}
		buf->ofs += received;
	}

	return res;
}

/* EOF */
