/* TTMPTestSvr.c
 *
 * TTMPServer test program.
 *
 * Copyright (C) 2005-2006 TomTom BV <http://www.tomtom.com/>
 * Author 0: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <stdio.h>
#include "TTMP.h"
#include "TTMPTestPlat.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

enum state_t {
	LISTENING,
	ACCEPTING,
	CONNECTED,
	RECEIVING,
};

static void rotate_slash(void)
{
	static const char slash[] = { '/', '-', '\\', '|' };
	static int state = 0;
	printf("\b%c", slash[state]);
	fflush(stdout);
	state = (state + 1) % sizeof slash;
}

static void print_error(const char *msg, int err)
{
	char buf[256];
	int res;

	res = TTMPErrorString(err, buf, sizeof buf);
	if (res != TTMP_ERR_NONE)
		printf("%s, error %d: Unable to get error string: error %d\n", msg, err, res);
	else
		printf("%s, error %d: %s\n", msg, err, buf);
}

int main(void)
{
	TTMPServer svr;
	int res;
	enum state_t state;
	TTMPConnection conn;
	TTMPMessage msg;

	signals_init();
	sockets_init();

	printf("Creating server...\n");
	res = TTMPServerCreate(&svr, 100000, VER_MAJOR, VER_MINOR);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to create server", res);
		goto error_exit;
	}

	printf("Listening on \"%s\"...\n", TEST_ADDRESS);
	res = TTMPServerListen(svr, TEST_ADDRESS);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to listen", res);
		goto error_destroy_svr;
	}
	state = LISTENING;

	for (;;) {
		switch (state) {
		case LISTENING:
			printf("Waiting for connection... ");
			fflush(stdout);
			state = ACCEPTING;
			/* FALLTHROUGH */
		case ACCEPTING:
			res = TTMPServerAccept(svr, &conn);
			if (res == TTMP_ERR_TIMEOUT) {
				rotate_slash();
				break;
			} else if (res == TTMP_ERR_INCOMPAT) {
				printf("\b\b\b\b, incompatible client.\n");
				state = LISTENING;
				break;
			} else if (res != TTMP_ERR_NONE) {
				print_error("\nUnable to accept", res);
				goto error_destroy_svr;
			}
			printf("\b\b\b\b, connected.\n");
			state = CONNECTED;
			/* FALLTHROUGH */
		case CONNECTED:
			printf("Receiving message... ");
			fflush(stdout);
			state = RECEIVING;
			/* FALLTHROUGH */
		case RECEIVING:
			res = TTMPConnectionReceive(conn, &msg);
			if (res == TTMP_ERR_TIMEOUT) {
				rotate_slash();
				break;
			} else if (res == TTMP_ERR_CONNEOF) {
				printf("\b\b\b\b, client left.\n");
				state = LISTENING;
				break;
			} else if (res != TTMP_ERR_NONE) {
				print_error("\nUnable to receive", res);
				goto error_destroy_svr;
			}
			printf("\b\b\b\b, done.\n");
			printf("Message contents:\n");
			TTMPMessageDump(msg, stdout);
			TTMPMessageDestroy(msg);
			state = CONNECTED;
			break;
		}
	}

	printf("\nDestroying server...\n");
	res = TTMPServerDestroy(svr);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy server", res);
		goto error_exit;
	}

	printf("Done.\n");
	return 0;

error_destroy_svr:
	printf("Destroying server...\n");
	res = TTMPServerDestroy(svr);
	if (res != TTMP_ERR_NONE) {
		print_error("Unable to destroy server", res);
	}

error_exit:
	return 1;
}

/* EOF */
